#include "sets.h"
#include "leveldb/options.h"

#include <chrono>
#include <print>


#include <leveldb/db.h>

#include <nlohmann/json.hpp>


using leveldb_ptr = leveldb::DB*;

auto sets_db     = leveldb_ptr{nullptr};
auto max_sets_id = 0;

extern "C"
{
    int open_sets_db()
    {
        auto opts = leveldb::Options{};

        opts.create_if_missing = true;

        auto status = leveldb::DB::Open(opts, "db/sets", &sets_db);

        if (!status.ok()) {
            std::println(stderr, "Failed to open sets database: {}", status.ToString());
            return 0;
        }

        auto it = sets_db->NewIterator(leveldb::ReadOptions());
        for (it->SeekToFirst(); it->Valid(); it->Next()) {
            auto id = std::stoi(it->key().ToString());
            if (id > max_sets_id) max_sets_id = id;
        }

        if (!it->status().ok()) {
            std::println(stderr, "Failed to get max set id: {}", it->status().ToString());
            delete it;
            delete sets_db;
            return 0;
        }

        delete it;
        return 1;
    }

    void close_sets_db()
    {
        delete sets_db;
    }

    int add_set(const char* name, const char* problems, const char* data, int* result)
    {
        auto json = nlohmann::json{
            {"name",     name                           },
            {"problems", nlohmann::json::parse(problems)},
            {"data",     nlohmann::json::parse(data)    }
        };
        auto status = sets_db->Put(leveldb::WriteOptions{}, std::to_string(max_sets_id + 1), json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to add a set: {}", status.ToString());
            return 0;
        }
        ++max_sets_id;
        *result = max_sets_id;
        return 1;
    }

    int delete_set(int id)
    {
        auto status = sets_db->Delete(leveldb::WriteOptions{}, std::to_string(id));
        if (status.IsNotFound()) {
            return -1;
        } else if (!status.ok()) {
            std::println(stderr, "Failed to delete a set: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    int get_set(int id, char** result)
    {
        auto value = std::string{};

        auto status = sets_db->Get(leveldb::ReadOptions{}, std::to_string(id), &value);
        if (status.IsNotFound()) {
            *result = NULL;
        } else if (!status.ok()) {
            std::println(stderr, "Failed to get a set: {}", status.ToString());
            return 0;
        } else {
            *result = strdup(value.c_str());
        }
        return 1;
    }

    int modify_set(int id, const char* name, const char* problems, const char* data)
    {
        auto json = nlohmann::json{
            {"name",     name                       },
            {"problems", nlohmann::json(problems)   },
            {"data",     nlohmann::json::parse(data)}
        };
        auto status = sets_db->Put(leveldb::WriteOptions{}, std::to_string(id), json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to add a set: {}", status.ToString());
            return 0;
        }
        return 1;
    }

    int all_sets(char** result)
    {
        auto problems = std::vector<int>{};
        auto buf      = 0;

        auto it = sets_db->NewIterator(leveldb::ReadOptions());
        for (it->SeekToFirst(); it->Valid(); it->Next()) {
            problems.emplace_back(std::stoi(it->key().ToString()));
        }

        if (!it->status().ok()) {
            std::println(stderr, "Failed to get all sets of the problem: {}", it->status().ToString());
            delete it;
            return 0;
        }
        delete it;
        auto json = nlohmann::json(problems);

        *result = strdup(json.dump().c_str());
        return 1;
    }
}