#include "hash.hpp"
#include "hash.h"

#include <cryptopp/algparam.h>
#include <cryptopp/cryptlib.h>
#include <cryptopp/filters.h>
#include <cryptopp/hex.h>
#include <cryptopp/modes.h>
#include <cryptopp/osrng.h>
#include <cryptopp/pwdbased.h>
#include <cryptopp/sha.h>

#include <array>
#include <format>
#include <random>
#include <vector>

using namespace CryptoPP;

auto to_byte_ptr(auto* p) noexcept -> const byte*
{
    return reinterpret_cast<const byte*>(p);
}

auto generate_salt(size_t length = 16) -> std::string
{
    auto prng = AutoSeededRandomPool{};
    auto salt = std::vector<byte>(length);
    prng.GenerateBlock(salt.data(), length);
    auto salt_str = std::string{};
    auto encoder  = HexEncoder{new StringSink(salt_str)};
    encoder.Put(salt.data(), length);
    encoder.MessageEnd();
    return salt_str;
}

auto hash_password_with_pbkdf2(const std::string_view password, const std::string_view salt, std::size_t iterations)
    -> std::string
{
    auto decoded_salt = std::string{};
    auto decoder      = HexDecoder{};
    decoder.Attach(new StringSink(decoded_salt));
    decoder.Put(to_byte_ptr(salt.data()), salt.size());
    decoder.MessageEnd();
    auto pbkdf2      = PKCS5_PBKDF2_HMAC<SHA256>{};
    auto derived_key = std::array<byte, SHA256::DIGESTSIZE>{};
    pbkdf2.DeriveKey(derived_key.data(),
                     derived_key.size(),
                     0,
                     to_byte_ptr(password.data()),
                     password.size(),
                     to_byte_ptr(decoded_salt.data()),
                     decoded_salt.size(),
                     iterations,
                     0);
    auto hashed_password = std::string{};
    auto encoder         = HexEncoder{new StringSink(hashed_password)};
    encoder.Put(derived_key.data(), derived_key.size());
    encoder.MessageEnd();
    return hashed_password;
}

auto generate_hash(const std::string_view password, std::size_t rounds) -> std::string
{
    auto salt            = generate_salt();
    auto hashed_password = hash_password_with_pbkdf2(password, salt, rounds);
    return std::format("{}${}${}", salt, rounds, hashed_password);
}

auto validate_password(const std::string_view password, const std::string_view hash) -> bool
{
    auto first_delimiter  = hash.find('$');
    auto second_delimiter = hash.find('$', first_delimiter + 1);

    if (first_delimiter == std::string::npos || second_delimiter == std::string::npos) {
        return false;
    }

    auto stored_salt   = hash.substr(0, first_delimiter);
    auto stored_rounds = std::size_t{};
    {
        auto first     = hash.data() + first_delimiter + 1;
        auto last      = hash.data() + second_delimiter;
        auto [ptr, ec] = std::from_chars(first, last, stored_rounds);
        if (ec != std::errc{}) return false;
    }
    auto stored_hashed_password = hash.substr(second_delimiter + 1);

    auto entered_hashed_password = hash_password_with_pbkdf2(password, stored_salt, stored_rounds);

    return entered_hashed_password == stored_hashed_password;
}

auto random_password(std::size_t length) -> std::string
{
    static const auto chars = std::string_view{"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"};

    auto rd  = std::random_device{};
    auto gen = std::mt19937{rd()};
    auto dis = std::uniform_int_distribution{std::size_t{0}, chars.size() - 1};

    std::string password;
    for (size_t i = 0; i < length; ++i) {
        password += chars[dis(gen)];
    }

    return password;
}


extern "C"
{
    char* kqm_generate_hash(const char* password, size_t rounds)
    {
        return strdup(generate_hash(password, rounds).c_str());
    }

    int kqm_validate_password(const char* password, const char* hash)
    {
        return validate_password(password, hash);
    }

    char* kqm_random_password(size_t length)
    {
        return strdup(random_password(length).c_str());
    }
}
